# Generated by asset_id_files.py

SIDEHOPPERSTATION_MREA = 0x02690DE1

NAME_TO_ID_MREA = {
    "01_SidehopperStation": 0x02690DE1,
}
# Generated by asset_id_files.py

SIDEHOPPERSTATION_MAPA = 0x2BBCF6DC

NAME_TO_ID_MAPA = {
    "01_SidehopperStation": 0x2BBCF6DC,
}

DOCK_NAMES = {
    "01_SidehopperStation": {
    },
}



# Generated by asset_id_files.py

CROSSFIRECHAOS_MREA = 0x44D34A2E

NAME_TO_ID_MREA = {
    "03_CrossfireChaos": 0x44D34A2E,
}
# Generated by asset_id_files.py

CROSSFIRECHAOS_MAPA = 0x6D06B113

NAME_TO_ID_MAPA = {
    "03_CrossfireChaos": 0x6D06B113,
}

DOCK_NAMES = {
    "03_CrossfireChaos": {
    },
}



import dataclasses

import construct
from open_prime_rando.echoes.asset_ids import world as world_ids
from open_prime_rando.echoes.inverted import area_pairs
from open_prime_rando.patcher_editor import PatcherEditor
from retro_data_structures.base_resource import Dependency, RawResource
from retro_data_structures.dependencies import recursive_dependencies_for_editor
from retro_data_structures.formats import Mlvl
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.script_layer import ScriptLayer
from retro_data_structures.formats.script_object import InstanceId, ScriptInstance
from retro_data_structures.properties.echoes.objects.AreaAttributes import AreaAttributes
from retro_data_structures.properties.echoes.objects.SafeZone import SafeZone
from retro_data_structures.properties.echoes.objects.SafeZoneCrystal import SafeZoneCrystal

_WORLDS = [
    world_ids.TEMPLE_GROUNDS_MLVL, world_ids.AGON_WASTES_MLVL, world_ids.TORVUS_BOG_MLVL,
    world_ids.SANCTUARY_FORTRESS_MLVL, world_ids.GREAT_TEMPLE_MLVL,
]
_AREAS_TO_SKIP = {
    0x775664a5,  # 00_scandummy
    0x752b2850,  # game_end_part1
    0x9d221f4a,  # game_end_part2
    0xc5250dbc,  # game_end_part3
    0x9641773f,  # game_end_part4
    0xce4665c9,  # game_end_part5
    0x4d4e5400,  # Menu
}


def collect_dependencies(obj):
    for field in dataclasses.fields(obj):
        if "asset_types" in field.metadata:
            yield getattr(obj, field.name)


_all_deps_cache: dict[frozenset[int], set[Dependency]] = {}


def _swap_dark_world(editor: PatcherEditor):
    for world_id in _WORLDS:
        world = editor.get_mlvl(world_id)

        for area in world.areas:
            if area.id in _AREAS_TO_SKIP:
                continue

            is_dark_world = area.internal_name.endswith("_dark")

            for layer in area.layers:
                for instance in layer.instances:
                    if instance.type == AreaAttributes:
                        prop = instance.get_properties()
                        assert isinstance(prop, AreaAttributes)
                        if prop.dark_world != is_dark_world:
                            print(area.name, is_dark_world, "found", prop.dark_world)
                        prop.dark_world = not is_dark_world
                        instance.set_properties(prop)

            mapa_id = world.mapw.get_mapa_id(area.index)
            mapa = bytearray(editor.get_raw_asset(mapa_id).data)
            mapa[8:12] = construct.Int32ub.build(not is_dark_world)
            editor.replace_asset(
                mapa_id,
                RawResource("MAPA", bytes(mapa))
            )


def _copy_safe_zones(dark: Area, dark_layer: ScriptLayer, light_layer: ScriptLayer,
                     new_dependencies: set[int], special_ids: set):
    copied_safe_zones = {}

    for instance in list(dark_layer.instances):
        if instance.type == SafeZone:
            copied_safe_zones[instance.id] = light_layer.add_instance_with(props := instance.get_properties())
            assert isinstance(props, SafeZone)
            new_dependencies.add(props.impact_effect)

            ids = {conn.target for conn in instance.connections}
            for target in ids:
                target_name = dark.get_instance(target).name
                if "ENTERED Safezone" not in target_name and "EXITED Safezone" not in target_name:
                    special_ids.add(instance.id)

            if instance.id not in special_ids:
                dark_layer.remove_instance(instance)

    return copied_safe_zones


def _copy_safe_zone_crystal(copied_crystal: ScriptInstance,
                            instance: ScriptInstance,
                            copied_safe_zones: dict[InstanceId, ScriptInstance],
                            dark: Area, special_ids: set[int],
                            ):
    targets_special = False

    for conn in instance.connections:
        if conn.target == instance.id:
            target = copied_crystal
        elif conn.target in copied_safe_zones:
            targets_special = targets_special or conn.target in special_ids
            target = copied_safe_zones[conn.target]
        else:
            targets_special = True
            weird_target = dark.get_instance(conn.target)
            print(
                f"Crystal {instance.name} at {dark.name} has unexpected connections "
                f"to {weird_target} ({weird_target.name})")
            continue
        copied_crystal.add_connection(conn.state, conn.message, target)

    return targets_special


def _update_dependencies(world: Mlvl, light: Area, new_dependencies: set[int]):
    # Add assets used by safe zones
    new_dependencies = frozenset(new_dependencies)
    if new_dependencies not in _all_deps_cache:
        _all_deps_cache[new_dependencies] = recursive_dependencies_for_editor(
            world.asset_manager, list(new_dependencies)
        )

    dependencies = light._raw.dependencies
    assert isinstance(dependencies.dependencies_b, list)
    for dep in _all_deps_cache[new_dependencies]:
        dependencies.dependencies_b.append(construct.Container(
            asset_id=dep.id,
            asset_type=dep.type,
        ))

    for i in range(1, len(dependencies.dependencies_offset)):
        dependencies.dependencies_offset[i] += len(new_dependencies)

    # Add Safe Zone's module dep
    module_dependencies = light._raw.module_dependencies
    for module_dep in set(SafeZone.modules() + SafeZoneCrystal.modules()):
        if module_dep not in module_dependencies.rel_module:
            module_dependencies.rel_module.insert(0, module_dep)
            for i in range(1, len(module_dependencies.rel_offset)):
                module_dependencies.rel_offset[i] += 1


def _move_safe_zones(world: Mlvl, pairs: list[tuple[int, int]]):
    for pair in pairs:
        light = world.get_area(pair[0])
        dark = world.get_area(pair[1])

        light_layer = light.get_layer("Default")
        dark_layer = dark.get_layer("Default")

        assert light_layer.index == 0

        new_dependencies: set[int] = set()  # asset ids that must be added to the mlvl dependency list
        special_ids = set()  # ids for objects that have extra functionality and we want to keep

        # Copy the safe zones
        copied_safe_zones = _copy_safe_zones(dark, dark_layer, light_layer, new_dependencies, special_ids)

        # Copy the safe zone crystals
        for instance in list(dark_layer.instances):
            if instance.type == SafeZoneCrystal:
                copied_crystal = light_layer.add_instance_with(props := instance.get_properties())
                new_dependencies.update(collect_dependencies(props))

                targets_special = _copy_safe_zone_crystal(
                    copied_crystal, instance, copied_safe_zones, dark, special_ids
                )
                if not targets_special:
                    dark_layer.remove_instance(instance)

        assert 0xffffffff not in new_dependencies
        # Update dependencies
        _update_dependencies(world, light, new_dependencies)


def apply_inverted(editor: PatcherEditor):
    _swap_dark_world(editor)

    _move_safe_zones(editor.get_mlvl(world_ids.TEMPLE_GROUNDS_MLVL), area_pairs.TG_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.AGON_WASTES_MLVL), area_pairs.AGON_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.TORVUS_BOG_MLVL), area_pairs.TORVUS_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.SANCTUARY_FORTRESS_MLVL), area_pairs.SANCTUARY_PAIRS)
    _move_safe_zones(editor.get_mlvl(world_ids.GREAT_TEMPLE_MLVL), area_pairs.GREAT_TEMPLE_PAIRS)



# Generated by asset_id_files.py

FRONTEND04_MREA = 0xCEE3AF0D

NAME_TO_ID_MREA = {
    "FrontEnd04": 0xCEE3AF0D,
}
# Generated by asset_id_files.py

FRONTEND04_MAPA = 0xE7365430

NAME_TO_ID_MAPA = {
    "FrontEnd04": 0xE7365430,
}

DOCK_NAMES = {
    "FrontEnd04": {
    },
}



# Generated by asset_id_files.py

MULTI_PIPELINE_MREA = 0x845DCF93

NAME_TO_ID_MREA = {
    "04_Multi_Pipeline": 0x845DCF93,
}
# Generated by asset_id_files.py

MULTI_PIPELINE_MAPA = 0xAD8834AE

NAME_TO_ID_MAPA = {
    "04_Multi_Pipeline": 0xAD8834AE,
}

DOCK_NAMES = {
    "04_Multi_Pipeline": {
    },
}



import dataclasses

from open_prime_rando.echoes.dock_lock_rando import dock_type
from open_prime_rando.echoes.dock_lock_rando.map_icons import DoorMapIcon
from open_prime_rando.echoes.vulnerabilities import resist_all_vuln, vulnerable
from retro_data_structures.properties.echoes.core.Color import Color
from retro_data_structures.properties.echoes.core.Vector import Vector

normal_door_model = 0x6B78FD92
dark_door_model = 0xbbcf134d
annihilator_door_model = 0xa50c7238

blast_collision_box = Vector(0.35, 5.0, 4.0)
blast_collision_offset = Vector(-2 / 3, 0, 2.0)

DOCK_TYPES: dict[str, dock_type.DoorType] = {
    "Normal": dock_type.NormalDoorType(
        name="Normal",
        vulnerability=dataclasses.replace(
            resist_all_vuln,
            power=vulnerable, dark=vulnerable, light=vulnerable, annihilator=vulnerable,
            power_charge=vulnerable, entangler=vulnerable, light_blast=vulnerable, sonic_boom=vulnerable,
            super_missle=vulnerable, black_hole=vulnerable, sunburst=vulnerable, imploder=vulnerable,

            missile=vulnerable, bomb=vulnerable, power_bomb=vulnerable,

        ),
        shell_color=Color(r=0, g=1, b=1, a=1),
        map_icon=DoorMapIcon.Normal,
    ),
    "Dark": dock_type.NormalDoorType(
        name="Dark",
        vulnerability=dataclasses.replace(resist_all_vuln, dark=vulnerable, entangler=vulnerable,
                                          black_hole=vulnerable),
        shell_model=dark_door_model,
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. Dark energy may damage it."
        ),
        map_icon=DoorMapIcon.Dark
    ),
    "Light": dock_type.NormalDoorType(
        name="Light",
        vulnerability=dataclasses.replace(resist_all_vuln, light=vulnerable, light_blast=vulnerable,
                                          sunburst=vulnerable),
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. Light energy may damage it."
        ),
        map_icon=DoorMapIcon.Light
    ),
    "Annihilator": dock_type.NormalDoorType(
        name="Annihilator",
        vulnerability=dataclasses.replace(resist_all_vuln, annihilator=vulnerable, sonic_boom=vulnerable,
                                          imploder=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A mix of light and dark energy may damage it."
        ),
        map_icon=DoorMapIcon.Annihilator
    ),
    "Disabled": dock_type.NormalDoorType(
        name="Disabled",
        vulnerability=resist_all_vuln,
        shell_color=Color(r=0, g=0, b=0, a=0),
        scan_text=(
            "Door system access denied.",
            "Unable to bypass security codes. Seek an alternate exit."
        ),
        map_icon=DoorMapIcon.Disabled
    ),
    "Missile": dock_type.VanillaBlastShieldDoorType(
        name="Missile",
        vulnerability=dataclasses.replace(resist_all_vuln, missile=vulnerable),
        shell_color=Color(r=1, g=0, b=0, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. A Missile blast may damage it."
        ),
        map_icon=DoorMapIcon.Missile,
        shield_model=0xBFB4A8EE,
    ),
    "SuperMissile": dock_type.VanillaBlastShieldDoorType(
        name="SuperMissile",
        vulnerability=dataclasses.replace(resist_all_vuln, super_missle=vulnerable),
        shell_color=Color(r=0, g=1, b=0, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A Super Missile blast may damage it."
        ),
        map_icon=DoorMapIcon.SuperMissile,
        shield_model=0xF115F575,
    ),
    "PowerBomb": dock_type.VanillaBlastShieldDoorType(
        name="PowerBomb",
        vulnerability=dataclasses.replace(resist_all_vuln, power_bomb=vulnerable),
        shell_color=Color(r=1, g=0.94, b=0, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. A Power Bomb may damage it."
        ),
        map_icon=DoorMapIcon.PowerBomb,
        shield_model=0xC2E4F075,
    ),
    "SeekerMissile": dock_type.SeekerBlastShieldDoorType(
        name="SeekerMissile",
        vulnerability=dataclasses.replace(resist_all_vuln, missile=vulnerable),
        shell_color=Color(r=0.5, g=0, b=0.64, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "Five simultaneous Missile blasts may damage it."
        ),
        map_icon=DoorMapIcon.SeekerMissile,
        shield_model=0x56F4208B,
    ),
    "ScrewAttack": dock_type.BlastShieldDoorType(
        name="ScrewAttack",
        vulnerability=dataclasses.replace(resist_all_vuln, screw_attack=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(r=0, g=0, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. The Screw Attack may damage it."
        ),
        map_icon=DoorMapIcon.ScrewAttack,
        shield_model=0x56F4208B,
    ),
    "Bomb": dock_type.BlastShieldDoorType(
        name="Bomb",
        vulnerability=dataclasses.replace(resist_all_vuln, bomb=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(r=0, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A Morph Ball Bomb blast may damage it."
        ),
        map_icon=DoorMapIcon.Bomb,
        shield_model=0x56F4208B,
    ),
    "Boost": dock_type.BlastShieldDoorType(
        name="Boost",
        vulnerability=dataclasses.replace(resist_all_vuln, boost_ball=vulnerable, cannon_ball=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(r=1, g=1, b=0, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. The Boost Ball may damage it."
        ),
        map_icon=DoorMapIcon.Boost,
        shield_model=0x56F4208B,
    ),
    "Grapple": dock_type.GrappleDoorType(
        name="Grapple",
        vulnerability=resist_all_vuln,
        shell_model=annihilator_door_model,
        shell_color=Color(r=1, g=0, b=0, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to weapon fire, but is weakly secured. "
            "The Grapple Beam may be able to remove it."
        ),
        map_icon=DoorMapIcon.Grapple,
        shield_model=0x56F4208B,
    ),
    "Darkburst": dock_type.BlastShieldDoorType(
        name="Darkburst",
        vulnerability=dataclasses.replace(resist_all_vuln, black_hole=vulnerable),
        shell_model=dark_door_model,
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of dark energy may damage it."
        ),
        map_icon=DoorMapIcon.Dark,
        shield_model="custom_door_lock_darkburst.CMDL",
    ),
    "Sunburst": dock_type.BlastShieldDoorType(
        name="Sunburst",
        vulnerability=dataclasses.replace(resist_all_vuln, sunburst=vulnerable),
        shell_model=normal_door_model,
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of light energy may damage it."
        ),
        map_icon=DoorMapIcon.Light,
        shield_model="custom_door_lock_sunburst.CMDL",
    ),
    "SonicBoom": dock_type.BlastShieldDoorType(
        name="SonicBoom",
        vulnerability=dataclasses.replace(resist_all_vuln, imploder=vulnerable),
        shell_model=annihilator_door_model,
        shell_color=Color(r=1, g=1, b=1, a=1),
        scan_text=(
            "There is a Blast Shield on the door blocking access. ",
            "Analysis indicates that the Blast Shield is invulnerable to most weapons. "
            "A massive burst of light and dark energy may damage it."
        ),
        map_icon=DoorMapIcon.Annihilator,
        shield_model="custom_door_lock_sonicboom.CMDL",
    ),
    "AgonEnergy": dock_type.PlanetaryEnergyDoorType(
        name="AgonEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(r=0.64, g=0.34, b=0, a=1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Agon. Return the energy to the Agon Temple."
        ),
        map_icon=DoorMapIcon.AgonEnergy,
        planetary_energy_item_id=-1  # TODO
    ),
    "TorvusEnergy": dock_type.PlanetaryEnergyDoorType(
        name="TorvusEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(r=0.31, g=0.59, b=38, a=1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Torvus. "
            "Return the energy to the Torvus Temple."
        ),
        map_icon=DoorMapIcon.TorvusEnergy,
        planetary_energy_item_id=-1  # TODO
    ),
    "SanctuaryEnergy": dock_type.PlanetaryEnergyDoorType(
        name="SanctuaryEnergy",
        vulnerability=resist_all_vuln,
        shell_color=Color(r=0.64, g=0.34, b=0, a=1),
        scan_text=(
            "There is a Luminoth barrier on the door blocking access. ",
            "Analysis indicates that the barrier is linked to the energy of Sanctuary. "
            "Return the energy to the Sanctuary Temple."
        ),
        map_icon=DoorMapIcon.AgonEnergy,
        planetary_energy_item_id=-1  # TODO
    ),
}



# Generated by asset_id_files.py

MULTI_SPIRES_MREA = 0x09749B2F

NAME_TO_ID_MREA = {
    "07_Multi_Spires": 0x09749B2F,
}
# Generated by asset_id_files.py

MULTI_SPIRES_MAPA = 0x20A16012

NAME_TO_ID_MAPA = {
    "07_Multi_Spires": 0x20A16012,
}

DOCK_NAMES = {
    "07_Multi_Spires": {
    },
}



import dataclasses
from typing import NamedTuple

from open_prime_rando.echoes.asset_ids import world
from open_prime_rando.echoes.dock_lock_rando.map_icons import DoorMapIcon
from open_prime_rando.echoes.vulnerabilities import resist_all_vuln
from open_prime_rando.patcher_editor import PatcherEditor
from retro_data_structures.asset_manager import NameOrAssetId
from retro_data_structures.base_resource import AssetId
from retro_data_structures.enums.echoes import Message, State, VisorFlags
from retro_data_structures.formats.mapa import Mapa
from retro_data_structures.formats.mrea import Area
from retro_data_structures.formats.scan import Scan
from retro_data_structures.formats.script_object import ScriptInstance
from retro_data_structures.formats.strg import Strg
from retro_data_structures.properties.base_property import BaseObjectType
from retro_data_structures.properties.echoes.archetypes.ActorParameters import ActorParameters
from retro_data_structures.properties.echoes.archetypes.DamageVulnerability import DamageVulnerability
from retro_data_structures.properties.echoes.archetypes.EditorProperties import EditorProperties
from retro_data_structures.properties.echoes.archetypes.HealthInfo import HealthInfo
from retro_data_structures.properties.echoes.archetypes.ScannableParameters import ScannableParameters
from retro_data_structures.properties.echoes.archetypes.SurroundPan import SurroundPan
from retro_data_structures.properties.echoes.archetypes.Transform import Transform
from retro_data_structures.properties.echoes.archetypes.VisorParameters import VisorParameters
from retro_data_structures.properties.echoes.archetypes.WeaponVulnerability import WeaponVulnerability
from retro_data_structures.properties.echoes.core.Color import Color
from retro_data_structures.properties.echoes.core.Spline import Spline
from retro_data_structures.properties.echoes.core.Vector import Vector
from retro_data_structures.properties.echoes.objects.Actor import Actor
from retro_data_structures.properties.echoes.objects.CameraShaker import CameraShaker
from retro_data_structures.properties.echoes.objects.Counter import Counter
from retro_data_structures.properties.echoes.objects.DamageableTrigger import DamageableTrigger
from retro_data_structures.properties.echoes.objects.DamageableTriggerOrientated import DamageableTriggerOrientated
from retro_data_structures.properties.echoes.objects.Dock import Dock
from retro_data_structures.properties.echoes.objects.Door import Door
from retro_data_structures.properties.echoes.objects.Effect import Effect
from retro_data_structures.properties.echoes.objects.MemoryRelay import MemoryRelay
from retro_data_structures.properties.echoes.objects.Relay import Relay
from retro_data_structures.properties.echoes.objects.ScannableObjectInfo import ScannableObjectInfo
from retro_data_structures.properties.echoes.objects.Sound import Sound
from retro_data_structures.properties.echoes.objects.StreamedAudio import StreamedAudio
from retro_data_structures.properties.echoes.objects.Timer import Timer


@dataclasses.dataclass(kw_only=True)
class DoorType:
    name: str

    vulnerability: DamageVulnerability

    shell_model: NameOrAssetId = 0x6B78FD92  # normal door model
    shell_color: Color

    scan_text: tuple[str, ...] | None = None

    map_icon: DoorMapIcon

    patched_scan: AssetId | None = None

    def get_paks(self, editor: PatcherEditor, world_name: str, area_name: str):
        world_file = world.load_dedicated_file(world_name)
        return editor.find_paks(world_file.NAME_TO_ID_MREA[area_name])

    def get_files(self, editor: PatcherEditor, world_name: str, area_name: str) -> tuple[Area, Mapa]:
        world_file = world.load_dedicated_file(world_name)
        area = editor.get_area(world.NAME_TO_ID_MLVL[world_name], world_file.NAME_TO_ID_MREA[area_name])
        mapa = editor.get_file(world_file.NAME_TO_ID_MAPA[area_name], type_hint=Mapa)
        return area, mapa

    def get_dock_index(self, world_name: str, area_name: str, dock_name: str) -> int:
        world_file = world.load_dedicated_file(world_name)
        return world_file.DOCK_NAMES[area_name][dock_name]

    def get_area(self, editor: PatcherEditor, world_name: str, area_name: str) -> Area:
        return self.get_files(editor, world_name, area_name)[0]

    def get_door_from_dock_index(self, area: Area, dock_index: int) -> ScriptInstance:
        dock = next(
            inst for inst in area.all_instances
            if (
                    inst.type == Dock and
                    inst.get_properties_as(Dock).dock_number == dock_index
            )
        )
        for instance in area.all_instances:
            if instance.type != Door:
                continue
            for connection in instance.connections:
                if dock.id_matches(connection.target):
                    return instance
        raise KeyError(f"no door found with a connection to {dock} in {area.name}")

    def patch_map_icon(self, mapa: Mapa, door: ScriptInstance):
        for obj in mapa.raw.mappable_objects:
            if door.id_matches(obj.editor_id):
                obj.type = self.map_icon.value
                return

    @staticmethod
    def get_scan_templates(editor: PatcherEditor) -> tuple[Scan, Strg]:
        # Uncategorized/There is a Blast Shield on the door blocking acces_0.SCAN
        scan = editor.get_parsed_asset(0x36DE1342, type_hint=Scan)
        # Strings/Uncategorized/There is a Blast Shield on the door blocking acces_0_0.STRG
        strg = editor.get_parsed_asset(0x49DF4448, type_hint=Strg)
        return scan, strg

    def get_patched_scan(self, editor: PatcherEditor, world_name: str, area_name: str) -> AssetId:
        paks = self.get_paks(editor, world_name, area_name)
        if self.patched_scan is None or not editor.does_asset_exists(self.patched_scan):
            scan, strg = DoorType.get_scan_templates(editor)
            for i, text in enumerate(self.scan_text):
                strg.set_string(i, text)

            strg_id = editor.add_or_replace_custom_asset(f"custom_door_{self.name}.STRG", strg, paks)

            with scan.scannable_object_info.edit_properties(ScannableObjectInfo) as scan_info:
                scan_info.string = strg_id

            scan_id = editor.add_or_replace_custom_asset(f"custom_door_{self.name}.SCAN", scan, paks)
            self.patched_scan = scan_id

        for pak in paks:
            editor.ensure_present(pak, self.patched_scan)
        return self.patched_scan

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        area, mapa = self.get_files(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))
        self.patch_map_icon(mapa, door)

        shell_model = editor._resolve_asset_id(self.shell_model)

        with door.edit_properties(Door) as door_props:
            door_props.shell_model = shell_model
            door_props.shell_color = self.shell_color
            if self.scan_text is not None:
                door_props.alt_scannable.scannable_info0 = self.get_patched_scan(editor, world_name, area_name)
            else:
                door_props.alt_scannable.scannable_info0 = 0xFFFFFFFF

        for pak in self.get_paks(editor, world_name, area_name):
            editor.ensure_present(pak, shell_model)


@dataclasses.dataclass(kw_only=True)
class NormalDoorType(DoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        with door.edit_properties(Door) as door_props:
            door_props.vulnerability = self.vulnerability


class BlastShieldActors(NamedTuple):
    door: ScriptInstance
    sound: ScriptInstance
    streamed: ScriptInstance
    lock: ScriptInstance
    relay: ScriptInstance
    gibs: ScriptInstance | None


@dataclasses.dataclass(kw_only=True)
class BlastShieldDoorType(DoorType):
    shield_model: NameOrAssetId
    shield_collision_box: Vector = dataclasses.field(default_factory=lambda: Vector(0.35, 5.0, 4.0))
    shield_collision_offset: Vector = dataclasses.field(default_factory=lambda: Vector(-2 / 3, 0, 2.0))

    def find_attached_instance(
            self, area: Area, source: ScriptInstance, state: State, message: Message,
            target_type: type[BaseObjectType], target_name: str | None = None
    ) -> ScriptInstance:
        for connection in source.connections:
            if connection.state == state and connection.message == message:
                target = area.get_instance(connection.target)
                if target.type == target_type and (target_name is None or target.name == target_name):
                    return target
        name = f"{target_type} named {target_name}" if target_name is not None else str(target_type)
        raise TypeError(f"No {name} connected to {source}")

    def get_spline(self) -> Spline:
        return Spline(
            data=b'\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x02'
                 b'\x02A \x00\x00?\x80\x00\x00\x02\x02\x01\x00\x00\x00\x00?\x80\x00\x00'
        )

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        """
        blast shield connections:
            DEAD -> lock cleared MemoryRelay, ACTV
            DEAD -> lock breaking SoundEffect, PLAY
            DEAD -> lock explosion gibs, ACTV
            DEAD -> jingle streamedaudio, PLAY

        door connections:
            OPEN -> lock cleared MemoryRelay, ACTV

        memory relay connections:
            ACTV -> door, RSET
            ACTV -> blast shield, DCTV
        """
        super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))
        door_transform = door.get_properties_as(Door).editor_properties.transform

        default = area.get_layer("Default")

        sound = default.add_instance_with(Sound(
            editor_properties=EditorProperties(
                name="Metal Door Lock Breaking",
                transform=door_transform
            ),
            sound=948,
            max_audible_distance=150.0,
            surround_pan=SurroundPan(surround_pan=1.0)
        ))

        streamed = default.add_instance_with(StreamedAudio(
            editor_properties=EditorProperties(
                name="StreamedAudio - Event Jingle",
                transform=door_transform
            ),
            song_file="/audio/evt_x_event_00.dsp",
            fade_in_time=0.01,
            volume=65,
            software_channel=1,
            unknown=False
        ))

        model = editor._resolve_asset_id(self.shield_model)
        lock = default.add_instance_with(Actor(
            editor_properties=EditorProperties(
                name=f"{self.name} Blast Shield Lock",
                transform=door_transform
            ),
            collision_box=self.shield_collision_box,
            collision_offset=self.shield_collision_offset,
            vulnerability=self.vulnerability,
            model=model,
            actor_information=ActorParameters(
                scannable=ScannableParameters(
                    scannable_info0=self.get_patched_scan(editor, world_name, area_name)
                )
            )
        ))

        relay = default.add_memory_relay("Lock Cleared")
        with relay.edit_properties(MemoryRelay) as _relay:
            _relay.editor_properties.transform = door_transform
            _relay.editor_properties.active = False

        lock.add_connection(State.Dead, Message.Play, sound)
        lock.add_connection(State.Dead, Message.Play, streamed)
        lock.add_connection(State.Dead, Message.Activate, relay)

        relay.add_connection(State.Active, Message.Deactivate, lock)
        relay.add_connection(State.Active, Message.Reset, door)

        door.add_connection(State.Open, Message.Activate, relay)

        dependencies = [
            model,
            0x8B4CD966,  # MetalDoorLockBreak AGSC
        ]

        gibs = None
        if not low_memory:
            particle = 0xCDCBDF04

            gibs = default.add_instance_with(Effect(
                editor_properties=EditorProperties(
                    transform=door_transform,
                    active=False
                ),
                particle_effect=particle,
                restart_on_activate=True,
                motion_control_spline=self.get_spline(),
            ))

            lock.add_connection(State.Dead, Message.Activate, gibs)
            dependencies.append(particle)

        for pak in self.get_paks(editor, world_name, area_name):
            for asset in dependencies:
                editor.ensure_present(pak, asset)

        return BlastShieldActors(door, sound, streamed, lock, relay, gibs)


@dataclasses.dataclass(kw_only=True)
class VanillaBlastShieldDoorType(BlastShieldDoorType):
    def remove_blast_shield(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str):
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        relay = self.find_attached_instance(area, door, State.Open, Message.Activate, MemoryRelay)
        lock = self.find_attached_instance(area, relay, State.Active, Message.Deactivate, Actor)

        for connection in lock.connections:
            area.remove_instance(connection.target)
        area.remove_instance(lock)


@dataclasses.dataclass(kw_only=True)
class SeekerBlastShieldDoorType(VanillaBlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        actors = super().patch_door(editor, world_name, area_name, dock_name, low_memory)
        area = self.get_area(editor, world_name, area_name)
        default = area.get_layer("Default")

        # immune instead of reflect so that it explodes on the lock
        missile_immune = dataclasses.replace(resist_all_vuln, missile=WeaponVulnerability(damage_multiplier=0.0))
        with actors.lock.edit_properties(Actor) as lock:
            lock.vulnerability = missile_immune
        with actors.door.edit_properties(Door) as door:
            door.vulnerability = missile_immune

        door_transform = actors.door.get_properties_as(Door).editor_properties.transform

        def create_trigger(name: str, health: float, lock_on: bool = True) -> ScriptInstance:
            pos = Vector(
                door_transform.position.x,
                door_transform.position.y,
                door_transform.position.z + 1.8
            )

            return default.add_instance_with(DamageableTriggerOrientated(
                editor_properties=EditorProperties(
                    name=name,
                    transform=Transform(
                        position=pos,
                        rotation=door_transform.rotation,
                        scale=Vector(4.0, 4.0, 1.5)
                    ),
                ),
                health=HealthInfo(health=health),
                vulnerability=self.vulnerability,
                enable_seeker_lock_on=lock_on,
                visor=VisorParameters(
                    visor_flags=VisorFlags.Combat | VisorFlags.Dark
                )
            ))

        timer = default.add_instance_with(Timer(
            editor_properties=EditorProperties(
                name="Button Control",
                transform=door_transform
            ),
            time=0.75
        ))

        timer_reset = default.add_instance_with(Timer(
            editor_properties=EditorProperties(
                name="Button Reset",
                transform=door_transform
            ),
            time=0.01
        ))

        # create 5 triggers so that you can have 5 lock-ons
        # 30.01 health because splash damage is inconsistent. missiles do 30 damage
        # so this guarantees you need at least 2 missiles at once to break it
        triggers = [create_trigger(f"Bridge Button {i}", 30.01) for i in range(5)]
        main_trigger = triggers[0]
        mini_trigger = create_trigger("Bridge Button Mini", 1.0, False)

        # start a timer when the tiny trigger dies. stop it if the main trigger dies
        mini_trigger.add_connection(State.Dead, Message.ResetAndStart, timer)
        main_trigger.add_connection(State.Dead, Message.Stop, timer)
        for trigger in triggers:
            timer.add_connection(State.Zero, Message.Deactivate, trigger)

        # if the timer counts all the way down, reset the triggers
        timer.add_connection(State.Zero, Message.ResetAndStart, timer_reset)
        timer_reset.add_connection(State.Zero, Message.Activate, mini_trigger)
        for trigger in triggers:
            timer_reset.add_connection(State.Zero, Message.Activate, trigger)

        # move the lock's connections to the trigger, since it's the thing that dies now
        for connection in actors.lock.connections:
            actors.lock.remove_connection(connection)
            main_trigger.add_connection(
                connection.state,
                connection.message,
                default.get_instance(connection.target)
            )

        for trigger in triggers:
            actors.relay.add_connection(State.Active, Message.Deactivate, trigger)
        actors.relay.add_connection(State.Active, Message.Deactivate, mini_trigger)
        actors.relay.add_connection(State.Active, Message.Deactivate, timer)
        actors.relay.add_connection(State.Active, Message.Deactivate, timer_reset)

    def remove_blast_shield(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str):
        area = self.get_area(editor, world_name, area_name)
        door = self.get_door_from_dock_index(area, self.get_dock_index(world_name, area_name, dock_name))

        memory_relay = self.find_attached_instance(area, door, State.Open, Message.Activate, MemoryRelay)
        trigger = self.find_attached_instance(area, memory_relay, State.Active, Message.Deactivate, DamageableTrigger)
        shaker = self.find_attached_instance(area, trigger, State.Dead, Message.Action, CameraShaker)
        counter = self.find_attached_instance(area, trigger, State.Dead, Message.Increment, Counter)
        complete_relay = self.find_attached_instance(area, counter, State.MaxReached, Message.SetToZero, Relay)

        area.remove_instance(shaker)
        for connection in complete_relay.connections:
            area.remove_instance(connection.target)
        area.remove_instance(complete_relay)
        area.remove_instance(counter)


@dataclasses.dataclass(kw_only=True)
class PlanetaryEnergyDoorType(DoorType):
    planetary_energy_item_id: int

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class GrappleDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class DarkVisorDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class EchoVisorDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class ScanVisorDoorType(BlastShieldDoorType):
    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()


@dataclasses.dataclass(kw_only=True)
class TranslatorDoorType(ScanVisorDoorType):
    translator_item_id: int

    def patch_door(self, editor: PatcherEditor, world_name: str, area_name: str, dock_name: str, low_memory: bool):
        raise NotImplementedError()



import random
from dataclasses import dataclass
from pathlib import Path

from open_prime_rando.echoes.asset_ids.sanctuary_fortress import MAIN_GYRO_CHAMBER_MREA
from open_prime_rando.echoes.asset_ids.world import SANCTUARY_FORTRESS_MLVL
from open_prime_rando.patcher_editor import PatcherEditor
from retro_data_structures.base_resource import RawResource
from retro_data_structures.enums.echoes import Message, State
from retro_data_structures.formats.cmdl import Cmdl
from retro_data_structures.properties.echoes.objects.Actor import Actor

RUBIKS_CUBES = {
    "Puzzle 1": [
        0x240013, 0x2401A4, 0x240145,
        0x24010A, 0x24000E, 0x240192,
        0x240139, 0x24013A, 0x240148,
    ],
    "Puzzle 2": [
        0x2402BA, 0x2402AD, 0x2402A6,
        0x2401C6, 0x2401BE, 0x2401C0,
        0x24028B, 0x2401C3, 0x240290,
    ]
}


@dataclass(frozen=True)
class RubiksColor:
    name: str
    state: State
    cmdl: int
    old_txtr: int

    @property
    def txtr_name(self) -> str:
        return f"rubiks_{self.name.lower()}.TXTR"

    @property
    def txtr(self) -> RawResource:
        return RawResource(
            type="TXTR",
            data=Path(__file__).parent.parent.joinpath("custom_assets", "rubiks", self.txtr_name).read_bytes()
        )


COLORS = {
    "RED": RubiksColor(
        name="red",
        state=State.InternalState00,
        cmdl=0xF21AB8BF,
        old_txtr=0x29A5FF4B,
    ),
    "GREEN": RubiksColor(
        name="green",
        state=State.InternalState01,
        cmdl=0x5F2ADFC3,
        old_txtr=0xEE2889A3,
    ),
    "BLUE": RubiksColor(
        name="blue",
        state=State.InternalState02,
        cmdl=0x8F25F715,
        old_txtr=0xFED23B81,
    )
}


def randomize_rubiks_puzzles(editor: PatcherEditor, rng: random.Random):
    area = editor.get_area(SANCTUARY_FORTRESS_MLVL, MAIN_GYRO_CHAMBER_MREA)

    # Add custom textures so colorblind players can distinguish the cubes
    for color in COLORS.values():
        txtr_id = editor.add_file(color.txtr_name, color.txtr, editor.find_paks(MAIN_GYRO_CHAMBER_MREA))

        cmdl = editor.get_file(color.cmdl, Cmdl)
        file_ids = cmdl.raw.material_sets[0].texture_file_ids
        old_txtr = file_ids.index(color.old_txtr)
        file_ids[old_txtr] = txtr_id

    for puzzle_name, cubes in RUBIKS_CUBES.items():
        solution = [
            COLORS["RED"], COLORS["RED"], COLORS["RED"],
            COLORS["GREEN"], COLORS["GREEN"], COLORS["GREEN"],
            COLORS["BLUE"], COLORS["BLUE"], COLORS["BLUE"],
        ]
        rng.shuffle(solution)

        puzzle = area.get_instance(puzzle_name)
        for color, cube_id in zip(solution, cubes):
            cube = area.get_instance(cube_id)

            puzzle.remove_connections_from(cube)
            puzzle.add_connection(color.state, Message.Attach, cube)

            with cube.edit_properties(Actor) as props:
                props.model = color.cmdl




from . import agon_wastes, great_temple, sanctuary_fortress, temple_grounds, torvus_bog, world

__all__ = [
    "agon_wastes", "great_temple", "sanctuary_fortress", "temple_grounds", "torvus_bog", "world",
]



# Generated by asset_id_files.py

AGON_WASTES_MLVL = 0x42B935E4
FRONTEND_MLVL = 0x69802220
GREAT_TEMPLE_MLVL = 0x863FCD72
M01_SIDEHOPPERSTATION_MLVL = 0xA50A80CC
M02_SPIRES_MLVL = 0xAE171602
M03_CROSSFIRECHAOS_MLVL = 0xE3B0C703
M04_PIPELINE_MLVL = 0x233E42BE
M05_SPIDERCOMPLEX_MLVL = 0x406ADD7F
M06_SHOOTINGGALLERY_MLVL = 0x7E19ED26
SANCTUARY_FORTRESS_MLVL = 0x1BAA96C2
TEMPLE_GROUNDS_MLVL = 0x3BFA3EFF
TORVUS_BOG_MLVL = 0x3DFD2249

NAME_TO_ID_MLVL = {
    "Agon Wastes": 0x42B935E4,
    "FrontEnd": 0x69802220,
    "Great Temple": 0x863FCD72,
    "M01_SidehopperStation": 0xA50A80CC,
    "M02_Spires": 0xAE171602,
    "M03_CrossfireChaos": 0xE3B0C703,
    "M04_Pipeline": 0x233E42BE,
    "M05_SpiderComplex": 0x406ADD7F,
    "M06_ShootingGallery": 0x7E19ED26,
    "Sanctuary Fortress": 0x1BAA96C2,
    "Temple Grounds": 0x3BFA3EFF,
    "Torvus Bog": 0x3DFD2249,
}
# Generated by asset_id_files.py

AGON_WASTES_MAPW = 0x38DBF6A5
FRONTEND_MAPW = 0x13E2E161
GREAT_TEMPLE_MAPW = 0xFC5D0E33
M01_SIDEHOPPERSTATION_MAPW = 0xDF68438D
M02_SPIRES_MAPW = 0xD475D543
M03_CROSSFIRECHAOS_MAPW = 0x99D20442
M04_PIPELINE_MAPW = 0x595C81FF
M05_SPIDERCOMPLEX_MAPW = 0x3A081E3E
M06_SHOOTINGGALLERY_MAPW = 0x047B2E67
SANCTUARY_FORTRESS_MAPW = 0x61C85583
TEMPLE_GROUNDS_MAPW = 0x4198FDBE
TORVUS_BOG_MAPW = 0x479FE108

NAME_TO_ID_MAPW = {
    "Agon Wastes": 0x38DBF6A5,
    "FrontEnd": 0x13E2E161,
    "Great Temple": 0xFC5D0E33,
    "M01_SidehopperStation": 0xDF68438D,
    "M02_Spires": 0xD475D543,
    "M03_CrossfireChaos": 0x99D20442,
    "M04_Pipeline": 0x595C81FF,
    "M05_SpiderComplex": 0x3A081E3E,
    "M06_ShootingGallery": 0x047B2E67,
    "Sanctuary Fortress": 0x61C85583,
    "Temple Grounds": 0x4198FDBE,
    "Torvus Bog": 0x479FE108,
}

_DEDICATED_FILES = {
    "Agon Wastes": ".agon_wastes",
    "FrontEnd": ".frontend",
    "Great Temple": ".great_temple",
    "M01_SidehopperStation": ".m01_sidehopperstation",
    "M02_Spires": ".m02_spires",
    "M03_CrossfireChaos": ".m03_crossfirechaos",
    "M04_Pipeline": ".m04_pipeline",
    "M05_SpiderComplex": ".m05_spidercomplex",
    "M06_ShootingGallery": ".m06_shootinggallery",
    "Sanctuary Fortress": ".sanctuary_fortress",
    "Temple Grounds": ".temple_grounds",
    "Torvus Bog": ".torvus_bog",
}


def load_dedicated_file(world_name: str):
    import importlib
    return importlib.import_module(
        _DEDICATED_FILES[world_name],
        ".".join(__name__.split(".")[:-1]),
    )



# Generated by asset_id_files.py

MULTI_SPIDERCOMPLEX_MREA = 0xE7095052

NAME_TO_ID_MREA = {
    "05_Multi_spiderComplex": 0xE7095052,
}
# Generated by asset_id_files.py

MULTI_SPIDERCOMPLEX_MAPA = 0xCEDCAB6F

NAME_TO_ID_MAPA = {
    "05_Multi_spiderComplex": 0xCEDCAB6F,
}

DOCK_NAMES = {
    "05_Multi_spiderComplex": {
    },
}



# Generated by asset_id_files.py

SHOOTINGGALLERY_MREA = 0xD97A600B

NAME_TO_ID_MREA = {
    "06_ShootingGallery": 0xD97A600B,
}
# Generated by asset_id_files.py

SHOOTINGGALLERY_MAPA = 0xF0AF9B36

NAME_TO_ID_MAPA = {
    "06_ShootingGallery": 0xF0AF9B36,
}

DOCK_NAMES = {
    "06_ShootingGallery": {
    },
}
